/** 

Remote controlled motorized roller curtain. 
Powered by a stepper motor, and controlled by an IR remote. 

by Axel Brinkeby
2018-05-02

Axel's website: http://axelsdiy.brinkeby.se/

compile for Arduino Nano. 

*/

#include <IRremote.h>   
// https://github.com/z3t0/Arduino-IRremote
// More info about library here: https://arduino-info.wikispaces.com/IR-RemoteControl

// pins:
const int LED1_PIN = A1; 
const int LED2_PIN = A0; 
const int IR_RX_PIN = 3;
const int STEPPER_EN_PIN = 4;
const int STEPPER_STEP_PIN = 5;
const int STEPPER_DIR_PIN = 6;
const int ENDSTOP_PIN = 2;
const int SOLENOID_PIN = 10;      // solenoid used to lock the motor with moving
const int BUTTON_UP_PIN = 7;
const int BUTTON_STOP_PIN = 8;
const int BUTTON_DOWN_PIN = 9;

// IR remote commands: 
// use the serial monitor to determine the codes for the buttons on the remote
const long UP_CMD = 0x6170D02F;         // UP button
const long STOP_CMD = 0x6170A857;       // OK button
const long DOWN_CMD = 0x617030CF;       // DOWN button

const long NUMBER_BTN_CMD[] = { 0x617000FF, 0x6170807F, 0x617040BF, 0x6170C03F, 0x617020DF, 
                                0x6170A05F, 0x6170609F, 0x6170E01F, 0x617010EF, 0x6170906F };  // btn 0 to 9

// the distance in (micro) steps the roller curtain shuld go down
const long DISTANCE = 28800; //200 * 16 * 9;    

// motor settings
const float ACCELERATION_VALUE = 0.03;    // 1 = no accel, almost 0 = very slow   
const long MOTOR_TIMER_VALUE = 58000;    // 1 - 65536  speed of the motor: higher value -> faster speed
const long MOTOR_START_GO_UP = 50;     // number of steps the motor goes up before going down, makes unlocking work better

IRrecv irrecv(IR_RX_PIN);
decode_results irData;

volatile long timerValue = MOTOR_TIMER_VALUE; 
volatile long targetPosition = 0; 
volatile long stepCounter = 0; 
volatile long startCounter = 0; 

bool btnState[] = {false, false, false}; 
bool prevBtnState[] = {false, false, false}; 

void setup() {
  Serial.begin(115200); // used for bebugging only. 

  pinMode(LED1_PIN, OUTPUT);            // status LEDs on the board
  pinMode(LED2_PIN, OUTPUT);  
  digitalWrite(LED1_PIN, false);
  digitalWrite(LED2_PIN, false);
  
  pinMode(STEPPER_EN_PIN, OUTPUT);      // for controlling the stepper motor driver
  pinMode(STEPPER_STEP_PIN, OUTPUT);  
  pinMode(STEPPER_DIR_PIN, OUTPUT);  
  digitalWrite(STEPPER_EN_PIN, true); 
  digitalWrite(STEPPER_STEP_PIN, false);
  digitalWrite(STEPPER_DIR_PIN, false);

  pinMode(BUTTON_UP_PIN, INPUT);        // buttons on the board
  pinMode(BUTTON_STOP_PIN, INPUT);  
  pinMode(BUTTON_DOWN_PIN, INPUT); 
  
  pinMode(ENDSTOP_PIN, INPUT);          // endstop at the top 
  pinMode(IR_RX_PIN, INPUT);            // reciever for IR remote control
  
  pinMode(SOLENOID_PIN, OUTPUT);        // solenoid used for locking the roller curtain in place while not moving
  digitalWrite(SOLENOID_PIN, false);

  irrecv.enableIRIn(); // Start the IR receiver

  // timer interrupt, 16 bit timer1 compare interrupt is used to drive the stepper motor
  TCCR1A = 0; 
  TCCR1B = 0;
  TCNT1 = MOTOR_TIMER_VALUE;
  TCCR1B |= (1<<CS10);  // prescaler
  TIMSK1 |= (1<<TOIE1); // TIMER OVERFLOW INTERRUPT
}

// 16 single cycle instructions is 1 microsecond at 16Mhz.
// used to make a step pulse to drive the motor. 
void delay_1us()
{
  __asm__ __volatile__ (
    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"
    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"
    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"
    "nop" "\n\t"    "nop" "\n\t"    "nop" "\n\t"    "nop");
}

// timer owerflow interrrupt, used to drive the motor. 
ISR(TIMER1_OVF_vect) {
  
  if (startCounter != 0 || targetPosition < stepCounter) {    // go up

    if (startCounter != 0) {
      startCounter--;
    }

    if (digitalRead(ENDSTOP_PIN)) {   // end stop hit, stop, and go back slightly. 
      stepCounter = -MOTOR_START_GO_UP;
      targetPosition = 0; 
    }
    
    digitalWrite(STEPPER_DIR_PIN, LOW); 
    digitalWrite(LED1_PIN, true);
    digitalWrite(LED2_PIN, false);
    digitalWrite(STEPPER_STEP_PIN, HIGH); 
    delay_1us();  
    digitalWrite(STEPPER_STEP_PIN, LOW);
    stepCounter--; 
    
  } else if (targetPosition > stepCounter) {     // go down
    digitalWrite(STEPPER_DIR_PIN, HIGH); 
    digitalWrite(LED1_PIN, false);
    digitalWrite(LED2_PIN, true);
    digitalWrite(STEPPER_STEP_PIN, HIGH); 
    delay_1us();  
    digitalWrite(STEPPER_STEP_PIN, LOW);
    stepCounter++;     
  }

  if (startCounter != 0) {
    // do some simple acceleration
    if (timerValue < MOTOR_TIMER_VALUE)
      timerValue += (MOTOR_TIMER_VALUE - timerValue) * ACCELERATION_VALUE;   
  }
  
  TCNT1 = timerValue; 
}

// set a position the motor shuld go to, the motor will 
// automaticly stop when the position is reached. 
void setTargetPosition(long pos) {
  targetPosition = pos;
  timerValue = MOTOR_TIMER_VALUE / 32;    // start the motor at a very slow speed
  startCounter = MOTOR_START_GO_UP; 
  digitalWrite(STEPPER_EN_PIN, false);
  unlock();   
}

// stop the motor at its current position
void stopMotor() {
  targetPosition = stepCounter;
}

// unlock the motor so it can move freely
void unlock() {
  digitalWrite(SOLENOID_PIN, true);    
}

// deactivetes the solenoid locking the motor
void lock() {  
  digitalWrite(SOLENOID_PIN, false); 
}

void loop() {
  // disable the stepper motor if it is not running. 
  if (targetPosition == stepCounter && !digitalRead(STEPPER_EN_PIN)) {
    digitalWrite(STEPPER_EN_PIN, true);
    lock();     
    digitalWrite(LED1_PIN, false);
    digitalWrite(LED2_PIN, false);
    Serial.println("Stop, motor disable and lock");
  }

  if (irrecv.decode(&irData)) {
    Serial.print(irData.value, HEX);   // print the value from the IR remote

    if (irData.value == UP_CMD) {       // make the roller curtain go all the way up to the end stop switch
      setTargetPosition(-DISTANCE);    
      Serial.print("  Going up");   
    }

    if (irData.value == STOP_CMD) {       // stop the roller curtain att the current position
      stopMotor();   
      Serial.print("  Stop");  
    }
    
    if (irData.value == DOWN_CMD) {       // make the roller curtain go all the way down 
      setTargetPosition(DISTANCE); 
      Serial.print("  Going down"); 
    }

    for(int i = 0; i < 10; i++) {
      if (irData.value == NUMBER_BTN_CMD[i]) {
        Serial.print("  Button number: "); 
        Serial.print(i); 
        int pos = DISTANCE * (float)(i / 10.0); 
        setTargetPosition(pos);
      }
    }

    Serial.println(); 
    
    irrecv.resume(); // Receive the next value
  }

  // read the states of the buttons, can also be used to control the roller curtain 
  // if you do not want to use IR remote control
  for (int i = 0; i < 3; i++) 
    prevBtnState[i] = btnState[i];     
  btnState[0] = !digitalRead(BUTTON_UP_PIN); 
  btnState[1] = !digitalRead(BUTTON_STOP_PIN); 
  btnState[2] = !digitalRead(BUTTON_DOWN_PIN); 

  if (btnState[0] && !prevBtnState[0]) {  // up button was just pressed down
    setTargetPosition(-DISTANCE);    
    Serial.println("Button: Up");
    delay(50); // to fix button debounce problem
  }

  if (btnState[1] && !prevBtnState[1]) {  // stop button was just pressed down
    stopMotor();   
    Serial.println("Button: Stop");
    delay(50); // to fix button debounce problem
  }

  if (btnState[2] && !prevBtnState[2]) {  // down button was just pressed down
    setTargetPosition(DISTANCE); 
    Serial.println("Button: Down"); 
    delay(50); // to fix button debounce problem
  }

}
